#!/bin/bash

MYPROJECT="myproject"

if [ ! -d $MYPROJECT ]; then
    echo "directory named '$MYPROJECT' not found"
    echo "execute:    mkdir $MYPROJECT"
    echo "and copy your project's files into the new directory"
    exit 1
fi

# --------------------------------------------------------------

TOPDIR=`pwd`
SAMPLEDIR="$TOPDIR/SAMPLE"
TESTDIR="$TOPDIR/TESTFILES"

TROVEFILE="/tmp/trove"

CSTD="c99"
#
MYEXE="./trove"
#
RUN4="$SAMPLEDIR/run4"
export RUN4_EXITSTATUS="/tmp/run4-exitstatus"
export RUN4_TIMEOUT="5sec"
#
cc -std=$CSTD -Wall -Werror -o $RUN4 $RUN4.c	|| exit 1
#
export RUN="/tmp/RUN"
if uname | grep -q Linux
then
    export PRIVATERUN="/tmp/RUN"
else
    export PRIVATERUN="/private/tmp/RUN"
fi

# --------------------------------------------------------------

if [ -t 1 ]; then
    RED="\033[1;31m" ; BLUE="\033[1;34m" ; OFF="\033[0m"
else
    RED="" ; BLUE="" ; OFF=""
fi

function red {
    echo -e "${RED}$1${OFF}"
}

function blue {
    echo -e "${BLUE}$1${OFF}"
}

function say {
    echo "        $1"
}

function dump_output {
    if [ -f $1 ]; then
	echo "        $2:"
	sed  's/^/            /' < $1
	echo
    fi
}

function addmark {
    MYTOTAL=`expr $MYTOTAL + $2`
    MAXIMUMTOTALMARK=`expr $MAXIMUMTOTALMARK + $3`
    if [ "$OUTCOME" = "Successful" ]
    then
	blue "    $1 - $OUTCOME ($2 / $3) -> ($MYTOTAL / $MAXIMUMTOTALMARK)"
    else
	red  "    $1 - $OUTCOME ($2 / $3) -> ($MYTOTAL / $MAXIMUMTOTALMARK)"
    fi
}

# --------------------------------------------------------------

function preclean {
    MARK="0"
    OUTCOME="Unsuccessful"
    rm -rf SO SE output? $TROVEFILE
}

# --------------------------------------------------------------

function test0 {
    preclean
    DESC="make compiled project without warnings or errors"
    echo "    ${FUNCNAME[0]} - $DESC"

    if [ -f Makefile ]; then
	say "compiling with provided Makefile"
	( /usr/bin/make -j1 > SO) &> SE
    else
	say "compiling with single cc invocation"
	(cc -std=$CSTD -Wall -Werror -o $MYEXE *.c > SO) &> SE
    fi

    if [ ! -f $MYEXE ]; then
	say "$MYEXE has not been created"
	exit 1
    fi

    if [ "$?" = "0" ]; then
	say "compilation successful"
	OUTCOME="Successful"
	MARK="$1"
	addmark ${FUNCNAME[0]} $MARK $1
    else
	red "compile failed"
	cat < SE
	OUTCOME="DNC"
    fi
}

# --------------------------------------------------------------

function test1 {
    preclean
    DESC="detects if no command-line arguments, prints error, terminates non-zero"
    echo "    ${FUNCNAME[0]} - $DESC"

    $RUN4 $MYEXE			&> output1
    ss=`cat $RUN4_EXITSTATUS`
    if [ "$ss" = "59" ]; then
	OUTCOME="TIMEOUT (after $RUN4_TIMEOUT)"
    else
	say "exited with $ss"
	if [ "$ss" != "0" ]; then	# exits with non-zero?
	    MARK=`expr $MARK + 1`
	fi
	if [ -s output1 ]; then		# prints an error message?
	    say "printed an error message"
	    MARK=`expr $MARK + 1`
	else
	    say "did not print an error message"
	fi
	if [ "$MARK" = "$1" ]; then	# fully successful?
	    OUTCOME="Successful"
	fi
    fi
    addmark ${FUNCNAME[0]} $MARK $1
}

# --------------------------------------------------------------

function test2 {
    preclean
    DESC="detects and reports unrecognised command-line argument"
    echo "    ${FUNCNAME[0]} - $DESC"

    $RUN4 $MYEXE -Q			&> output1
    ss=`cat $RUN4_EXITSTATUS`
    if [ "$ss" = "59" ]; then
	OUTCOME="TIMEOUT (after $RUN4_TIMEOUT)"
    else
	say "exited with $ss"
	if [ "$ss" != "0" ]; then	# exits with non-zero?
	    MARK=`expr $MARK + 1`
	fi
	if [ -s output1 ]; then		# prints an error message?
	    say "printed an error message"
	    MARK=`expr $MARK + 1`
	else
	    say "did not print an error message"
	fi
	if [ "$MARK" = "$1" ]; then	# fully successful?
	    OUTCOME="Successful"
	fi
    fi
    addmark ${FUNCNAME[0]} $MARK $1
}

# --------------------------------------------------------------

function test3 {
    preclean
    DESC="detects non-existent file in filelist, prints error, terminates non-zero"
    echo "    ${FUNCNAME[0]} - $DESC"

    $RUN4 $MYEXE -b FILE-DOES-NOT-EXIST	&> output1
    ss=`cat $RUN4_EXITSTATUS`
    if [ "$ss" = "59" ]; then
	OUTCOME="TIMEOUT (after $RUN4_TIMEOUT)"
    else
	say "exited with $ss"
	if [ "$ss" != "0" ]; then	# exits with non-zero?
	    MARK=`expr $MARK + 1`
	fi
	if [ -s output1 ]; then		# prints an error message?
	    say "printed an error message"
	    MARK=`expr $MARK + 1`
	else
	    say "did not print an error message"
	fi
	if [ "$MARK" = "$1" ]; then	# fully successful?
	    OUTCOME="Successful"
	fi
    fi
    addmark ${FUNCNAME[0]} $MARK $1
}

# --------------------------------------------------------------

function test4 {
    preclean
    DESC="builds /tmp/trove file from a single file"
    echo "    ${FUNCNAME[0]} - $DESC"

    $RUN4 $MYEXE -b $TESTDIR/baseball	&> /dev/null
    ss=`cat $RUN4_EXITSTATUS`
    if [ "$ss" = "59" ]; then
	OUTCOME="TIMEOUT (after $RUN4_TIMEOUT)"
    else
	say "exited with $ss"
	if [ "$ss" = "0" ]; then	# exits with zero?
	    MARK=`expr $MARK + 1`
	fi
	if [ -f $TROVEFILE ]; then	# produces a trove file
	    say "produced $TROVEFILE"
	    ls -l $TROVEFILE | sed 's/^/        /'
	    MARK=`expr $MARK + 1`
	else
	    say "did not produce $TROVEFILE"
	fi
	if [ "$MARK" = "$1" ]; then	# fully successful?
	    OUTCOME="Successful"
	fi
    fi
    addmark ${FUNCNAME[0]} $MARK $1
}

# --------------------------------------------------------------

function test5 {
    preclean
    DESC="supports -f option to build trove file with different name"
    echo "    ${FUNCNAME[0]} - $DESC"
    MYTROVE="/tmp/budgie"

    rm -rf $MYTROVE
    $RUN4 $MYEXE -f $MYTROVE -b $TESTDIR/baseball	&> /dev/null
    ss=`cat $RUN4_EXITSTATUS`
    if [ "$ss" = "59" ]; then
	OUTCOME="TIMEOUT (after $RUN4_TIMEOUT)"
    else
	say "exited with $ss"
	if [ "$ss" = "0" ]; then	# exits with zero?
	    MARK=`expr $MARK + 1`
	fi
	if [ -f $MYTROVE ]; then	# produces a different trove file
	    say "produced $MYTROVE"
	    ls -l $MYTROVE | sed 's/^/        /'
	    MARK=`expr $MARK + 1`
	else
	    say "did not produce $MYTROVE"
	fi
	if [ "$MARK" = "$1" ]; then	# fully successful?
	    OUTCOME="Successful"
	fi
    fi
    addmark ${FUNCNAME[0]} $MARK $1
}

# --------------------------------------------------------------

function test6 {
    preclean
    DESC="builds a trove file from a filelist containing files and directories"
    echo "    ${FUNCNAME[0]} - $DESC"

    $RUN4 $MYEXE -b $TESTDIR/rotate* $TESTDIR/textfiles		&> /dev/null
    ss=`cat $RUN4_EXITSTATUS`
    if [ "$ss" = "59" ]; then
	OUTCOME="TIMEOUT (after $RUN4_TIMEOUT)"
    else
	say "exited with $ss"
	if [ "$ss" = "0" ]; then	# exits with zero?
	    MARK=`expr $MARK + 1`
	fi
	if [ -f $TROVEFILE ]; then	# produces a trove file
	    say "produced $TROVEFILE"
	    ls -l $TROVEFILE | sed 's/^/        /'
	    MARK=`expr $MARK + 1`
	else
	    say "did not produce $TROVEFILE"
	fi
	if [ "$MARK" = "$1" ]; then	# fully successful?
	    OUTCOME="Successful"
	fi
    fi
    addmark ${FUNCNAME[0]} $MARK $1
}

# --------------------------------------------------------------

function test7 {
    preclean
    DESC="builds a trove file from a single textfile, finds a word from that file"
    echo "    ${FUNCNAME[0]} - $DESC"
    WORD="return"

    $RUN4 $MYEXE -b $TESTDIR/rotate.c	&> /dev/null
    ss=`cat $RUN4_EXITSTATUS`
    if [ "$ss" = "59" ]; then
	OUTCOME="TIMEOUT (after $RUN4_TIMEOUT)"
    else
	if [ -f $TROVEFILE ]; then	# produces default trove file
	    say "produced $TROVEFILE"
	    ls -l $TROVEFILE | sed 's/^/        /'

	    $RUN4 $MYEXE $WORD		&> output2
	    ss=`cat $RUN4_EXITSTATUS`
	    if [ "$ss" = "59" ]; then
		OUTCOME="TIMEOUT (after $RUN4_TIMEOUT)"
	    else
		say "exited with $ss"
		if grep -q '^/.*rotate.c$' < output2
		then
		    say "found required word '$WORD' in output"
		    MARK="$1"
		    OUTCOME="Successful"
		else
		    say "did not find required word '$WORD' in output"
		fi
	    fi
	fi
    fi
    addmark ${FUNCNAME[0]} $MARK $1
}

# --------------------------------------------------------------

function test8 {
    preclean
    DESC="builds a trove file from filelist, finds multiple instances of a word"
    echo "    ${FUNCNAME[0]} - $DESC"
    WORD="return"
    TIMES="8"

    $RUN4 $MYEXE -b $TESTDIR		&> /dev/null
    ss=`cat $RUN4_EXITSTATUS`
    if [ "$ss" = "59" ]; then
	OUTCOME="TIMEOUT (after $RUN4_TIMEOUT)"
    else
	say "exited with $ss"
	if [ -f $TROVEFILE ]; then	# produces default trove file

	    $RUN4 $MYEXE $WORD		&> output2
	    ss=`cat $RUN4_EXITSTATUS`
	    if [ "$ss" = "59" ]; then
		OUTCOME="TIMEOUT (after $RUN4_TIMEOUT)"
	    else
		say "exited with $ss"
		if grep '^/' < output2 | wc -l | grep -q $TIMES
		then
		    say "found required word output"
		    MARK="$1"
		    OUTCOME="Successful"
		else
		    say "did not find required word in output"
		fi
	    fi
	fi
    fi
    addmark ${FUNCNAME[0]} $MARK $1
}

# --------------------------------------------------------------

function test9 {
    preclean
    DESC="builds a trove file from a filelist, removes one of the files"
    echo "    ${FUNCNAME[0]} - $DESC"
    WORD="day1"
    TIMES="1"

    cp $TESTDIR/testsrc/mycal/*	.
    $RUN4 $MYEXE -b first_day_of_month.c mycal.c solution.c	&> /dev/null
    ss=`cat $RUN4_EXITSTATUS`
    if [ "$ss" = "59" ]; then
	OUTCOME="TIMEOUT (after $RUN4_TIMEOUT)"
    else
	say "exited with $ss"
	if [ -f $TROVEFILE ]; then	# produces default trove file

	    $RUN4 $MYEXE -r $PRIVATERUN/first_day_of_month.c	&> /dev/null
	    ss=`cat $RUN4_EXITSTATUS`
	    if [ "$ss" = "59" ]; then
		OUTCOME="TIMEOUT (after $RUN4_TIMEOUT)"
	    else
		say "exited with $ss"
		$RUN4 $MYEXE $WORD	&> output3
		ss=`cat $RUN4_EXITSTATUS`
		if [ "$ss" = "59" ]; then
		    OUTCOME="TIMEOUT (after $RUN4_TIMEOUT)"
		else
		    say "exited with $ss"
		    if grep '^/' < output3 | wc -l | grep -q $TIMES
		    then
			say "found required word output"
			MARK="$1"
			OUTCOME="Successful"
		    else
			say "did not find required word in output"
		    fi
		fi
	    fi
	fi
    fi
    addmark ${FUNCNAME[0]} $MARK $1
    rm -f first_day_of_month.c mycal.c solution.c
}

# --------------------------------------------------------------

function testA {
    preclean
    DESC="builds a trove file from a filelist, updates one of the files"
    echo "    ${FUNCNAME[0]} - $DESC"
    WORD="volume"
    TIMES="3"

    cp $TESTDIR/baseball $TESTDIR/pizza $TESTDIR/rotate.c $TESTDIR/csseduck.png .
    $RUN4 $MYEXE -b baseball pizza rotate.c csseduck.png	&> /dev/null
    ss=`cat $RUN4_EXITSTATUS`
    if [ "$ss" = "59" ]; then
	OUTCOME="TIMEOUT (after $RUN4_TIMEOUT)"
    else
	say "exited with $ss"
	if [ -f $TROVEFILE ]; then	# produces default trove file

	    echo '// adding volume' >> rotate.c
	    $RUN4 $MYEXE -u rotate.c		&> /dev/null
	    ss=`cat $RUN4_EXITSTATUS`
	    if [ "$ss" = "59" ]; then
		OUTCOME="TIMEOUT (after $RUN4_TIMEOUT)"
	    else
		say "exited with $ss"
		$RUN4 $MYEXE $WORD	&> output3
		ss=`cat $RUN4_EXITSTATUS`
		if [ "$ss" = "59" ]; then
		    OUTCOME="TIMEOUT (after $RUN4_TIMEOUT)"
		else
		    say "exited with $ss"
		    if grep '^/' < output3 | wc -l | grep -q $TIMES
		    then
			say "found required word in output"
			MARK="$1"
			OUTCOME="Successful"
		    else
			say "did not find required word in output"
		    fi
		fi
	    fi
	fi
    fi
    addmark ${FUNCNAME[0]} $MARK $1
    rm -f baseball pizza rotate.c csseduck.png
}

# --------------------------------------------------------------

function testB {
    preclean
    DESC="trove file is a file compressed by gzip"
    echo "    ${FUNCNAME[0]} - $DESC"

    $RUN4 $MYEXE -b $TESTDIR/baseball	&> /dev/null
    ss=`cat $RUN4_EXITSTATUS`
    if [ "$ss" = "59" ]; then
	OUTCOME="TIMEOUT (after $RUN4_TIMEOUT)"
    else
	say "exited with $ss"
	if [ -f $TROVEFILE ]; then	# produces a trove file
	    if file $TROVEFILE | grep -q 'gzip compressed'
	    then
		say "$TROVEFILE appears to be a file compressed with gzip"
		MARK="$1"
		OUTCOME="Successful"
	    else
		say "$TROVEFILE does not appear to be a file compressed with gzip"
	    fi
	fi
    fi
    addmark ${FUNCNAME[0]} $MARK $1
}

# --------------------------------------------------------------

#  IN $TOPDIR DIRECTORY WHEN CALLED
function prepare_automated_tests {
#  COPY SUBMITTED FILES TO $RUN,  APPEND newlines TO SOURCE FILES
    cp -pR $MYPROJECT/* $RUN
    for f in $RUN/*.[ch]
    do
	echo >> $f
    done
}

#  IN POPULATED $RUN DIRECTORY WHEN CALLED
function automated_tests {
    MYTOTAL="0"
    MAXIMUMTOTALMARK="0"

    test0 3
    if [ "$OUTCOME" == "DNC" ] ; then
	red "compile failed"
	$MAXIMUMTOTALMARK="25"
    else
	test1  2
	test2  2
	test3  2
	test4  2
	test5  2
	test6  2
	test7  2
	test8  2
	test9  2
	testA  2
	testB  2
    fi
}

# --------------------------------------------------------------

rm -rf $RUN ; mkdir $RUN
prepare_automated_tests
cd $RUN
automated_tests
blue "MARK FROM AUTOMATED TESTS: $MYTOTAL / $MAXIMUMTOTALMARK"

rm -rf $RUN $RUN4_EXITSTATUS

